
CONLL_STRUCTURE = ['id', 'orth', 'lemma', 'pos', 'pos2', 'morph', 'parent', 'relation', '_1', '_2', 'as' ]




# class with up and down traverse methods and token requirement checking
class Sentence:

    def parseConllSent(self, conllSent):
        byId = {}
        for line in conllSent:
            # read line:
            for idx, val in enumerate(line.strip().split('\t')):
                if CONLL_STRUCTURE[idx] == 'id':
                    id = int(val)
                    byId[id] = {}
                if CONLL_STRUCTURE[idx] == 'parent':
                    byId[id][CONLL_STRUCTURE[idx]] = int(val)
                else:
                    byId[id][CONLL_STRUCTURE[idx]] = val
        return byId

    def __init__(self, conllSent):
        self.byId = self.parseConllSent( conllSent )
        self.ids = self.byId.keys()

    # here we assume it's the last column
    def getTokenIDsWithAttribute(self, attribute):
        return [Id for Id, tokenDict in self.byId.items() if tokenDict['as']==attribute ]
    

    # check node requirements (as specified in [..] )
    def checkNodeRequirements(self, nodeId, requirementsList):
        if type(nodeId) is list or nodeId==0:
            print "checkNodeRequirements requires int not list:",nodeId
            return None
        for attribute,value in requirementsList.items():
            if not self.byId[nodeId][attribute] == value:
                return False
        return True


    # easy, as each token has only one parent
    def nodesUpFrom(self, nodeId, relation=None):
        # no relation defined, just return parent node
        if relation is None:
            return self.byId[nodeId]['parent']
        # relation defined, check if matches:
        elif self.byId[nodeId]['relation'] == relation:
            return self.byId[nodeId]['parent'] # yes, return parent it
        else:
            return None


    # move up from the nodes we're/might be in, return a collection of possible ids
    def moveUp(self, currentPossibleNodeIds, visitedNodeIds, requiredLabel=None):
        possibleIds = []
        for nodeId in currentPossibleNodeIds:
            nodeup = self.nodesUpFrom(nodeId, requiredLabel)
            if nodeup is not None and nodeup not in visitedNodeIds:
                possibleIds.append( nodeup )
        #return list of possible node ids
        return possibleIds


    def moveDown(self, currentPossibleNodeIds, visitedNodeIds, requiredLabel=None):
        possibleIds = []
        for nodeId in currentPossibleNodeIds:
            # unfortunately have to review all nodes for their parents:
            for childId in self.ids:
                if self.byId[childId]['parent'] == nodeId:
                    if requiredLabel is not None:
                        if self.byId[childId]['relation'] == requiredLabel and childId not in visitedNodeIds:
                            possibleIds.append(childId)
                    else:
                        possibleIds.append(childId)

        #return list of possible node ids
        return possibleIds



__author__ = 'aleks'

import pyparsing
from pyparsing import Word, Literal, alphanums, ZeroOrMore, Suppress, Optional, Group, oneOf
import pickle
from sentence import *

class SentencePatternMatcher():

    equals = Literal('=')
    word = Word(alphanums)
    tagAndVal = Group(word.setResultsName("tag") + Suppress(":") +word.setResultsName("val")).setResultsName("tagAndVal")  + Suppress(Optional(";"))
    token = Suppress("[") + Group(ZeroOrMore(tagAndVal)).setResultsName("tagAndVals")  +  Suppress("]")

    edge = oneOf(["<", ">"])+word.setResultsName("label")

    # parse [tag1:val1;tag2:val2] and return python dictionary
    def parseNode(self, tokenString):
        try:
            result = self.token.parseString(tokenString)
        except pyparsing.ParseException:
            print "Path parsing exception: Malformed token string: ",tokenString
        requirementsDict = {}
        #print "result:",result
        for r in result.tagAndVals:
            #print "r:",r
            #print "tagAndVal.tag=",r.tag
            #print "tagAndVal.val=",r.val
            requirementsDict[r.tag] = r.val
        return requirementsDict
        #print "tagAndVal/s:",result.tagAndVals.tagAndVal

    # return label from edge definition
    def parseTag(self, tagString):
        try:
            result = self.edge.parseString(tagString)
        except pyparsing.ParseException:
            print "Path parsing exception: Malformed edge string: ",tagString
        return result.label


    # check sentence against pattern
    def matchPathInSentence(self, startNodeIds, sgpath, s):

        tokens = sgpath.split(' ')
        visitedNodeIds = currentPossibleNodeIds = startNodeIds
        match = 1

        for token in tokens:

            # we are now at token:
            if token.startswith("["):
                requirementsList = self.parseNode(token)
                # check all possible current nodes for complying with requirements:
                validIds = []
                for nodeId in currentPossibleNodeIds:
                    if s.checkNodeRequirements(nodeId, requirementsList):
                        validIds.append(nodeId)
                # no node meets those requirements:
                if len(validIds) == 0:
                    match=0 ; break
                else:
                    currentPossibleNodeIds = validIds
                    visitedNodeIds.extend( currentPossibleNodeIds )

            # we are now at edge indicating upward
            elif token.startswith("<"):
                requiredLabel = self.parseTag(token)
                # get nodes with this label:
                currentNodeIds = s.moveUp(currentPossibleNodeIds, visitedNodeIds, requiredLabel)
                if currentNodeIds is None:  # label did not match
                    match=0 ; break
                else:
                    currentPossibleNodeIds = currentNodeIds

            # we are now at edge indicating downward
            elif token.startswith(">"):
                requiredLabel = self.parseTag(token)
                currentNodeIds = s.moveDown(currentPossibleNodeIds, visitedNodeIds, requiredLabel)
                if currentNodeIds is None:  # label did not match
                    match=0 ; break
                else:
                    currentPossibleNodeIds = currentNodeIds

        if match==0:
            currentPossibleNodeIds = None

        return currentPossibleNodeIds

